"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .itempathjson import ItemPathJSON
import datetime as dt
from .datavaluejson import DataValueJSON
from .epmvariable import EpmVariable
from .epmproperty import EpmProperty
from .dataobjectattributes import DataObjectAttributes
from .epmnodeids import EpmNodeIds
from .epmdataobject import EpmDataObject
from .basicvariablepropertymask import BasicVariablePropertyMask


class BasicVariable(EpmDataObject):
    """description of class"""

    def __init__(self, epmConnection, identity, name, description = None, tagType = None, realTimeEnabled = None, deadBandFilter = None, 
                 deadBandUnit = None, eu = None, lowLimit = None, highLimit = None, scaleEnable = None, inputLowLimit = None, 
                 inputHighLimit = None, clamping = None, domain = None, interface = None, ioTagAddress = None, processingEnabled = None, 
                 isRecording = None, isCompressing = None, storeMillisecondsEnabled = None, storageSet = None):
        super().__init__(epmConnection, name, identity)
        self._name = name
        self._newName = None
        self._description = description
        self._tagType = tagType
        self._realTimeEnabled = realTimeEnabled
        self._deadBandFilter = deadBandFilter
        self._deadBandUnit = deadBandUnit
        self._eu = eu
        self._lowLimit = lowLimit
        self._highLimit = highLimit
        self._scaleEnable = scaleEnable
        self._inputLowLimit = inputLowLimit
        self._inputHighLimit = inputHighLimit
        self._clamping = clamping
        self._domain = domain
        self._interface = interface
        self._ioTagAddress = ioTagAddress
        self._processingEnabled = processingEnabled
        self._isRecording = isRecording
        self._isCompressing = isCompressing
        self._storeMillisecondsEnabled = storeMillisecondsEnabled
        self._storageSet = storageSet

    @property
    def name(self):
      return self._name

    @name.setter
    def name(self, value):
      if self._name == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Name.value
      self._newName = value

    @property
    def tagType(self):
      return self._tagType

    @tagType.setter
    def tagType(self, value):
      if self._tagType == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.TagType.value
      self._tagType = value

    @property
    def realTimeEnabled(self):
      return self._realTimeEnabled

    @realTimeEnabled.setter
    def realTimeEnabled(self, value):
      if self._realTimeEnabled == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.RealTimeEnabled.value
      self._realTimeEnabled = value

    @property
    def deadBandFilter(self):
      return self._deadBandFilter

    @deadBandFilter.setter
    def deadBandFilter(self, value):
      if self._deadBandFilter == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.DeadBandFilter.value
      self._deadBandFilter = value
        
    @property
    def deadBandUnit(self):
      return self._deadBandUnit

    @deadBandUnit.setter
    def deadBandUnit(self, value):
      if self._deadBandUnit == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.DeadBandUnit.value
      self._deadBandUnit = value

    @property
    def scaleEnable(self):
      return self._scaleEnable

    @scaleEnable.setter
    def scaleEnable(self, value):
      if self._scaleEnable == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.ScaleEnable.value
      self._scaleEnable = value

    @property
    def inputLowLimit(self):
      return self._inputLowLimit

    @inputLowLimit.setter
    def inputLowLimit(self, value):
      if self._inputLowLimit == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.InputLowLimit.value
      self._inputLowLimit = value
        
    @property
    def inputHighLimit(self):
      return self._inputHighLimit

    @inputHighLimit.setter
    def inputHighLimit(self, value):
      if self._inputHighLimit == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.InputHighLimit.value
      self._inputHighLimit = value
        
    @property
    def interface(self):
      return self._interface

    @interface.setter
    def interface(self, value):
      if self._interface == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Interface.value
      self._interface = value
        
    @property
    def ioTagAddress(self):
      return self._ioTagAddress

    @ioTagAddress.setter
    def ioTagAddress(self, value):
      if self._ioTagAddress == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.IoTagAddress.value
      self._ioTagAddress = value
        
    @property
    def processingEnabled(self):
      return self._processingEnabled

    @processingEnabled.setter
    def processingEnabled(self, value):
      if self._processingEnabled == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.ProcessingEnabled.value
      self._processingEnabled = value
        
    @property
    def isRecording(self):
      return self._isRecording

    @isRecording.setter
    def isRecording(self, value):
      if self._isRecording == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.IsRecording.value
      self._isRecording = value
        
    @property
    def isCompressing(self):
      return self._isCompressing

    @isCompressing.setter
    def isCompressing(self, value):
      if self._isCompressing == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.IsCompressing.value
      self._isCompressing = value
        
    @property
    def storeMillisecondsEnabled(self):
      return self._storeMillisecondsEnabled 

    @storeMillisecondsEnabled.setter
    def storeMillisecondsEnabled(self, value):
      if self._storeMillisecondsEnabled == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.StoreMillisecondsEnabled.value
      self._storeMillisecondsEnabled = value
        
    @property
    def storageSet(self):
      return self._storageSet

    @storageSet.setter
    def storageSet(self, value):
      if self._storageSet == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.StorageSet.value
      self._storageSet = value

    #Public Methods

    def save(self):
      self._epmConnection.updateBasicVariable(self._name, 
                                              self._newName if self._hasFlag(BasicVariablePropertyMask.Name) else None,
                                              self._description if self._hasFlag(BasicVariablePropertyMask.Description) else None,
                                              self._tagType if self._hasFlag(BasicVariablePropertyMask.TagType) else None,
                                              self._realTimeEnabled if self._hasFlag(BasicVariablePropertyMask.RealTimeEnabled) else None,
                                              self._deadBandFilter if self._hasFlag(BasicVariablePropertyMask.DeadBandFilter) else None,
                                              self._deadBandUnit if self._hasFlag(BasicVariablePropertyMask.DeadBandUnit) else None,
                                              self._eu if self._hasFlag(BasicVariablePropertyMask.Eu) else None,
                                              self._lowLimit if self._hasFlag(BasicVariablePropertyMask.LowLimit) else None,
                                              self._highLimit if self._hasFlag(BasicVariablePropertyMask.HighLimit) else None,
                                              self._scaleEnable if self._hasFlag(BasicVariablePropertyMask.ScaleEnable) else None,
                                              self._inputLowLimit if self._hasFlag(BasicVariablePropertyMask.InputLowLimit) else None,
                                              self._inputHighLimit if self._hasFlag(BasicVariablePropertyMask.InputHighLimit) else None,
                                              self._clamping if self._hasFlag(BasicVariablePropertyMask.Clamping) else None,
                                              self._domain if self._hasFlag(BasicVariablePropertyMask.Domain) else None,
                                              self._interface if self._hasFlag(BasicVariablePropertyMask.Interface) else None,
                                              self._ioTagAddress if self._hasFlag(BasicVariablePropertyMask.IoTagAddress) else None,
                                              self._processingEnabled if self._hasFlag(BasicVariablePropertyMask.ProcessingEnabled) else None,
                                              self._isRecording if self._hasFlag(BasicVariablePropertyMask.IsRecording) else None,
                                              self._isCompressing if self._hasFlag(BasicVariablePropertyMask.IsCompressing) else None,
                                              self._storeMillisecondsEnabled if self._hasFlag(BasicVariablePropertyMask.StoreMillisecondsEnabled) else None,
                                              self._storageSet if self._hasFlag(BasicVariablePropertyMask.StorageSet) else None)

      if (self._newName != None and self._hasFlag(BasicVariablePropertyMask.Name)):
        self._name = self._newName

      self._changeMask = BasicVariablePropertyMask.Unspecified.value

      


    def copy(self, newName, description = None, tagType = None, realTimeEnabled = None, deadBandFilter = None, 
            deadBandUnit = None, eu = None, lowLimit = None, highLimit = None, scaleEnable = None, inputLowLimit = None, 
            inputHighLimit = None, clamping = None, domain = None, interface = None, ioTagAddress = None, processingEnabled = None, 
            isRecording = None, isCompressing = None, storeMillisecondsEnabled = None, storageSet = None):
      return self._epmConnection.createBasicVariable(newName,
                                              description = description if description != None else self._description,
                                              tagType = tagType if tagType != None else self._tagType,
                                              realTimeEnabled = realTimeEnabled if realTimeEnabled != None else self._realTimeEnabled,
                                              deadBandFilter = deadBandFilter if deadBandFilter != None else self._deadBandFilter,
                                              deadBandUnit = deadBandUnit if deadBandUnit != None else self._deadBandUnit,
                                              eu = eu if eu != None else self._eu,
                                              lowLimit = lowLimit if lowLimit != None else self._lowLimit,
                                              highLimit = highLimit if highLimit != None else self._highLimit,
                                              scaleEnable = scaleEnable if scaleEnable != None else self._scaleEnable,
                                              inputLowLimit = inputLowLimit if inputLowLimit != None else self._inputLowLimit,
                                              inputHighLimit = inputHighLimit if inputHighLimit != None else self._inputHighLimit,
                                              clamping = clamping if clamping != None else self._clamping,
                                              domain = domain if domain != None else self._domain,
                                              interface = interface if interface != None else self._interface,
                                              ioTagAddress = ioTagAddress if ioTagAddress != None else self._ioTagAddress,
                                              processingEnabled = processingEnabled if processingEnabled != None else self._processingEnabled,
                                              isRecording = isRecording if isRecording != None else self._isRecording,
                                              isCompressing = isCompressing if isCompressing != None else self._isCompressing,
                                              storeMillisecondsEnabled = storeMillisecondsEnabled if storeMillisecondsEnabled != None else self._storeMillisecondsEnabled,
                                              storageSet = storageSet if storageSet != None else self._storageSet)

    def delete(self):
      result = self._epmConnection.deleteBasicVariable([ self._name ])
      return result[0]

    def _hasFlag(self, flag):
      return self._changeMask & flag.value == flag.value



class BasicVariableAlreayExistsException(Exception):
    pass

class BasicVariableInvalidNameException(Exception):
    pass

class StorageSetDoesNotExistException(Exception):
    pass

class InterfaceDoesNotExistException(Exception):
    pass




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class QueryResultMask(Enum):

    Unknown = 0

    Identity = 1

    Name = 2

    Description = 4

    Type = 8

    All = 31




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class AnnotationValueJSON(object):
    """description of class"""
    def __init__(self, message, userName, annotationTime):
        self._message = message
        self._userName = userName
        self._annotationTime = annotationTime

    @property
    def message(self):
      return self._message

    @property
    def userName(self):
      return self._userName

    @property
    def annotationTime(self):
      return self._annotationTime

    def toDict(self):
        return {'value': { 
                            'annotationTime' : self._annotationTime.isoformat(), 
                            'message' : self._message,
                            'userName' : self._userName 
                         }, 
                'quality': 0, 
                'timestamp' : self._annotationTime.isoformat() }



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class DownloadType(Enum):
    Text = "Text";
    Json = "Json";
    Binary = "Binary"







"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class HistoryUpdateResultModelJSON(object):
    """description of class"""
    def __init__(self, operationResults, diagnostics):
        self._operationResults = operationResults
        self._diagnostics = diagnostics





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import IntFlag

class DataObjectAttributes(IntFlag):

    Unspecified = 0

    Name = 1

    Description = 2

    EU = 4

    LowLimit = 8

    HighLimit = 16

    Clamping = 32

    Domain = 64

    All = 127




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import json
class QueryModelJSON(object):
    """description of class"""
    def __init__(self, continuationPoint, release, resultMask, browseNameFilter, descriptionFilter, euNameFilter, typeFilter, domainFilter):
      self._continuationPoint = continuationPoint
      self._release = release
      self._resultMask = resultMask
      self._browseNameFilter = browseNameFilter
      self._descriptionFilter = descriptionFilter
      self._euNameFilter = euNameFilter
      self._typeFilter = typeFilter
      self._domainFilter = domainFilter

    def toDict(self):
        jsonPaths = []

        types = []
        for item in self._typeFilter:
            types.append(item.toDict());

        return {'continuationPoint': self._continuationPoint, 'release': self._release, 
               'resultMask' : self._resultMask, 'browseNameFilter' : self._browseNameFilter, 
               'descriptionFilter' : self._descriptionFilter, 'euNameFilter' : self._euNameFilter, 
               'typeFilter' : types, 'domainFilter' : self._domainFilter }




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class AggregateType(Enum):
    # <summary>
    # The identifier for the Trend Object.
    # </summary>
    Trend = 'Trend'

    # <summary>
    # The identifier for the Interpolative Object.
    # </summary>
    Interpolative = 'Interpolative'

    # <summary>
    # The identifier for the Average Object.
    # </summary>
    Average = 'Average'

    # <summary>
    # The identifier for the TimeAverage Object.
    # </summary>
    TimeAverage = 'TimeAverage'

    # <summary>
    # The identifier for the Total Object.
    # </summary>
    Total = 'Total'

    # <summary>
    # The identifier for the Minimum Object.
    # </summary>
    Minimum = 'Minimum'

    # <summary>
    # The identifier for the Maximum Object.
    # </summary>
    Maximum = 'Maximum'

    # <summary>
    # The identifier for the MinimumActualTime Object.
    # </summary>
    MinimumActualTime = 'MinimumActualTime'

    # <summary>
    # The identifier for the MaximumActualTime Object.
    # </summary>
    MaximumActualTime = 'MaximumActualTime'

    # <summary>
    # The identifier for the Range Object.
    # </summary>
    Range = 'Range'

    # <summary>
    # The identifier for the AnnotationCount Object.
    # </summary>
    AnnotationCount = 'AnnotationCount'

    # <summary>
    # The identifier for the Count Object.
    # </summary>
    Count = 'Count'

    # <summary>
    # The identifier for the DurationInStateZero Object.
    # </summary>
    DurationInStateZero = 'DurationInStateZero'

    # <summary>
    # The identifier for the DurationInStateNonZero Object.
    # </summary>
    DurationInStateNonZero = 'DurationInStateNonZero'

    # <summary>
    # The identifier for the PercentInStateZero Object.
    # </summary>
    PercentInStateZero = 'PercentInStateZero'

    # <summary>
    # The identifier for the PercentInStateNonZero Object.
    # </summary>
    PercentInStateNonZero = 'PercentInStateNonZero'

    # <summary>
    # The identifier for the NumberOfTransitions Object.
    # </summary>
    NumberOfTransitions = 'NumberOfTransitions'

    # <summary>
    # The identifier for the Start Object.
    # </summary>
    Start = 'Start'

    # <summary>
    # The identifier for the End Object.
    # </summary>
    End = 'End'

    # <summary>
    # The identifier for the Delta Object.
    # </summary>
    Delta = 'Delta'

    # <summary>
    # The identifier for the DurationGood Object.
    # </summary>
    DurationGood = 'DurationGood'

    # <summary>
    # The identifier for the DurationBad Object.
    # </summary>
    DurationBad = 'DurationBad'

    # <summary>
    # The identifier for the PercentGood Object.
    # </summary>
    PercentGood = 'PercentGood'

    # <summary>
    # The identifier for the PercentBad Object.
    # </summary>
    PercentBad = 'PercentBad'

    # <summary>
    # The identifier for the WorstQuality Object.
    # </summary>
    WorstQuality = 'WorstQuality'

    # <summary>
    # The identifier for the TimeAverage2 Object.
    # </summary>
    TimeAverage2 = 'TimeAverage2'

    # <summary>
    # The identifier for the Total2 Object.
    # </summary>
    Total2 = 'Total2'

    # <summary>
    # The identifier for the Minimum2 Object.
    # </summary>
    Minimum2 = 'Minimum2'

    # <summary>
    # The identifier for the Maximum2 Object.
    # </summary>
    Maximum2 = 'Maximum2'

    # <summary>
    # The identifier for the MinimumActualTime2 Object.
    # </summary>
    MinimumActualTime2 = 'MinimumActualTime2'

    # <summary>
    # The identifier for the MaximumActualTime2 Object.
    # </summary>
    MaximumActualTime2 = 'MaximumActualTime2'

    # <summary>
    # The identifier for the Range2 Object.
    # </summary>
    Range2 = 'Range2'

    # <summary>
    # The identifier for the StartBounds Object.
    # </summary>
    StartBounds = 'StartBounds'

    # <summary>
    # The identifier for the EndBounds Object.
    # </summary>
    EndBounds = 'EndBounds'

    # <summary>
    # The identifier for the DeltaBounds Object.
    # </summary>
    DeltaBounds = 'DeltaBounds'

    # <summary>
    # The identifier for the WorstQuality2 Object.
    # </summary>
    WorstQuality2 = 'WorstQuality2'

    # <summary>
    # The identifier for the StandardDeviationPopulation Object.
    # </summary>
    StandardDeviationPopulation = 'StandardDeviationPopulation'

    # <summary>
    # The identifier for the VariancePopulation Object.
    # </summary>
    VariancePopulation = 'VariancePopulation'

    # <summary>
    # The identifier for the StandardDeviationSample Object.
    # </summary>
    StandardDeviationSample = 'StandardDeviationSample'

    # <summary>
    # The identifier for the VarianceSample Object.
    # </summary>
    VarianceSample = 'VarianceSample'

class AggregateDetails(object):
    """description of class"""

    def __init__(self, interval, type):
        self._interval = interval
        self._type = type.value

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, value):
        self._interval = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class BrowseItemModelJSON(object):
  
  def __init__(self, path, nodeClassMask, referenceTypes):
    self._path = path
    self._nodeClassMask = nodeClassMask
    self._referenceTypes = referenceTypes

  def toDict(self):

    referenceTypes = []
    for item in self._referenceTypes:
        referenceTypes.append(item);

    return { 'path': self._path.toDict(), 
             'nodeClassMask': self._nodeClassMask, 
             'referenceTypes': referenceTypes }




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from threading import Timer, Thread, Event

class AutoStartTimer(object):
  def __init__(self, spam, callback):
    self._spam = spam
    self._callback = callback
    self._thread = Timer(self._spam, self.callback)
    self._thread.start()

  def callback(self):
    self._callback()
    self._thread = Timer(self._spam, self.callback)
    self._thread.start()

  def start(self):
    self._thread.start()

  def cancel(self):
    self._thread.cancel()



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import requests
import json
import collections

from .folder import Folder
from .resource import Resource
from .downloadtype import DownloadType

from .mimetype import Application, Text, Image, Audio, Video

class ResourcesManager(object):

  def __init__(self, authorizationService, webApi, baseFolderId):
    self._authorizationService = authorizationService
    self._webApi = webApi
    self._baseFolderId = baseFolderId

  def browse(self):
    return self._browse(self._baseFolderId)

  def getResource(self, path):
    return self._getResource(self._baseFolderId + (('/' + path) if len(path) > 0 else ''))

  def createFolder(self, name, description=None):
    return self._createFolder(self._baseFolderId, name, description)

  def _getResource(self, path = None, id = None):

    odata = ""
    if path != None:
      odata = "files('path:" + path + "')"
    elif id != None:
      odata = "files('" + id + "')"

    token = self._authorizationService.getToken()
    
    url = self._webApi + "/resource/odata/v2/" + odata
    header = {'Authorization': 'Bearer {}'.format(token)}

    response = requests.get(url, headers=header, verify=False)
    if response.status_code != 200:
        print(response.reason)
        raise Exception("Browse service call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)

    json_result = json.loads(response.text)
    if json_result == None:
      raise Exception("Read Failed no result")

    if json_result['mimeType'] == Application.ElipsePortalFolder.value:
      return Folder(self, json_result['identity'], json_result['name'], json_result['description'])
    else:
      return Resource(self, json_result['identity'], json_result['name'], json_result['description'], json_result['mimeType'])

  def _browse(self, folderId):

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/odata/v2/files('" + folderId + "')/children(recursive=false)?$format=application/json;odata.metadata=none"
    header = {'Authorization': 'Bearer {}'.format(token)}

    response = requests.get(url, headers=header, verify=False)
    if response.status_code != 200:
        print(response.reason)
        raise Exception("Browse service call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)

    json_result = json.loads(response.text)
    if json_result == None:
        raise Exception("Read Failed no result")

    children = collections.OrderedDict()
    for item in json_result['value']:
      if item['mimeType'] == Application.ElipsePortalFolder.value:
        children[item['name']] = Folder(self, item['identity'], item['name'], item['description'])
      else:
        children[item['name']] = Resource(self, item['identity'], item['name'], item['description'], item['mimeType'])

    return children

  def _download(self, resourceId, type):

    import urllib
    from urllib.request import urlopen, Request
    import io
    from shutil import copyfileobj

    if isinstance(type, DownloadType):
      typeName = type.value
    else:
      typeName = type

    resourceId = urllib.parse.quote(resourceId, safe='')

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/v2/files/" + resourceId
    header = {'Authorization': 'Bearer {}'.format(token)}

    request = Request(url, headers=header)

    chunk_size = 8096
    with urlopen(request) as response:
      if typeName == DownloadType.Binary.value:
        resourceStream = io.BytesIO()
        copyfileobj(response, resourceStream, chunk_size)
        resourceStream.seek(0)
        return resourceStream
      elif typeName == DownloadType.Text.value:
        return response.read().decode('utf-8')
      elif typeName == DownloadType.Json.value:
        return json.loads(response.read())
      else:
        raise Exception("Invalid download file type '" +  str(response.status_code) + "'. Reason: " + response.reason)

  def _upload(self, parentId, name, file, description = None, mimeType = None, thumbnail = None, thumbnailMimeType = None, starred = None, overrideFile = False):

    if isinstance(mimeType, Application) | isinstance(mimeType, Text) | isinstance(mimeType, Image) | isinstance(mimeType, Audio) | isinstance(mimeType, Video):
      mimeTypeName = mimeType.value
    else:
      mimeTypeName = mimeType


    from requests_toolbelt.multipart.encoder import MultipartEncoder

    metadata = { 'name': name,
                 'description': description,
                 'mimeType': mimeTypeName,
                 'thumbnail': thumbnail,
                 'thumbnailMimeType': thumbnailMimeType,
                 'parents': [ parentId],
                 'starred': starred }

    m = MultipartEncoder(fields={'__metadata': ( "blob", json.dumps(metadata), "application/json"), 
                                 'file': (name, file, mimeTypeName)})

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/v2/upload/files"
    header = {'Content-Type': m.content_type, 'Authorization': 'Bearer {}'.format(token)}

    response = requests.post(url, headers=header, data=m, verify=False)

    if response.status_code == 409 and overrideFile:
      exitentResource = self._browse(parentId)[name]
      file.seek(0)
      self._updateFile(exitentResource._id, exitentResource._name, exitentResource._mimeType, file)
      return exitentResource

    elif response.status_code != 200:
      print(response.reason)
      raise Exception("Upload call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)

    json_result = json.loads(response.text)

    resource =  Resource(self, json_result['items'][0]['identity'], json_result['items'][0]['name'], json_result['items'][0]['description'], json_result['items'][0]['mimeType'])

    return resource

  def _createFolder(self, parentId, name, description):

    from requests_toolbelt.multipart.encoder import MultipartEncoder

    metadata = { 'name': name,
                 'description': description,
                 'mimeType': 'application/vnd.elipse.portal.folder',
                 'thumbnail': None,
                 'thumbnailMimeType': None,
                 'parents': [ parentId ],
                 'starred': None }

    m = MultipartEncoder(fields={'__metadata': ( "blob", json.dumps(metadata), "application/json")})

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/v2/upload/files"
    header = {'Content-Type': m.content_type, 'Authorization': 'Bearer {}'.format(token)}

    response = requests.post(url, headers=header, data=m, verify=False)

    if response.status_code != 200:
      print(response.reason)
      raise Exception("CreateFolder call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)
    json_result = json.loads(response.text)

    folder = Folder(self, json_result['items'][0]['identity'], json_result['items'][0]['name'], json_result['items'][0]['description'])

    return folder

  def _updateFile(self, id, name, mimeType, file):

    from requests_toolbelt.multipart.encoder import MultipartEncoder
    import urllib

    metadata = { 'name': name,
                 #'description': description,
                 'mimeType': mimeType }
                 #'thumbnail': thumbnail,
                 #'thumbnailMimeType': thumbnailMimeType,
                 #'parents': [ parentId],
                 #'starred': starred }

    m = MultipartEncoder(fields={'__metadata': ( "blob", json.dumps(metadata), "application/json"), 
                                 'file': (name, file, mimeType)})

    id = urllib.parse.quote(id, safe='')

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/v2/files/" + id
    header = {'Content-Type': m.content_type, 'Authorization': 'Bearer {}'.format(token)}

    response = requests.patch(url, headers=header, data=m, verify=False)

    if response.status_code != 200:
      print(response.reason)
      raise Exception("Upload call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)

  def _delete(self, id):

    token = self._authorizationService.getToken()
    url = self._webApi + "/resource/odata/v2/files('" + id + "')"
    header = {'Authorization': 'Bearer {}'.format(token)}

    response = requests.delete(url, headers=header, verify=False)
    if response.status_code != 200:
        print(response.reason)
        raise Exception("Delete service call http error '" +  str(response.status_code) + "'. Reason: " + response.reason)




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""
import pytz

from .itempathjson import ItemPathJSON
import datetime as dt
from .queryperiod import QueryPeriod
from .datavaluejson import DataValueJSON
from .epmvariable import EpmVariable
from .epmproperty import EpmProperty
from .dataobjectattributes import DataObjectAttributes
from .epmnodeids import EpmNodeIds
from .basicvariablepropertymask import BasicVariablePropertyMask

from enum import Enum
import collections


class EpmDataObjectPropertyNames(Enum):
    Name = "4:Name"

    Description = "4:Description"

    EU = "0:EngineeringUnits"

    LowLimit = "4:RangeLow"

    HighLimit = "4:RangeHigh"

    Clamping = "4:RangeClamping"

    Domain = "4:Discrete"

    Annotations = "4:Annotations"


class EpmDataObjectAttributeIds(Enum):
    Name = 1

    Description = 2

    LowLimit = 15

    HighLimit = 16

    Clamping = 17

    Domain = 21

    EU = 22

    Annotations = 900


class RetrievalMode(Enum):
    Previous = "Previous"

    Exact = "Exact"

    Next = "Next"


def getDiscreteValue(domain):
  if domain == 'Discrete':
    return True
  elif domain == 'Continuous':
    return False
  else:
    return None

def getDomainValue(discrete):
  if discrete == None:
    return None
  elif discrete:
    return 'Discrete'
  else:
    return 'Continuous'

class EpmDataObject(EpmVariable):
    """description of class"""

    def __init__(self, epmConnection, name, identity):
        super().__init__(epmConnection, name, '/DataObjects/' + name, identity)
        self._changeMask = BasicVariablePropertyMask.Unspecified.value
        self._identity = identity
        self._description = None
        self._eu = None
        self._highLimit = None
        self._lowLimit = None
        self._clamping = None
        self._domain = None

    ### Properties
    @property
    def description(self):
      return self._description

    @description.setter
    def description(self, value):
      if self._description == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Description.value
      self._description = value

    @property
    def eu(self):
      return self._eu

    @eu.setter
    def eu(self, value):
      if self._eu == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Eu.value
      self._eu = value

    @property
    def lowLimit(self):
      return self._lowLimit

    @lowLimit.setter
    def lowLimit(self, value):
      if self._lowLimit == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.LowLimit.value
      self._lowLimit = value

    @property
    def highLimit(self):
      return self._lowLimit

    @highLimit.setter
    def highLimit(self, value):
      if self._highLimit == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.HighLimit.value
      self._highLimit = value

    @property
    def clamping(self):
      return self._clamping

    @clamping.setter
    def clamping(self, value):
      if self._clamping == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Clamping.value
      self._clamping = value

    @property
    def domain(self):
      return self._domain

    @domain.setter
    def domain(self, value):
      if self._domain == value:
        return
      self._changeMask = self._changeMask | BasicVariablePropertyMask.Domain.value
      self._domain = value

    ## Public Methods
    def recordedValue(self, timestamp, retrieval = RetrievalMode.Previous):
        timestamp = timestamp.astimezone(pytz.UTC)
        start = timestamp - dt.timedelta(milliseconds=1)
        end = timestamp + dt.timedelta(milliseconds=1)
        queryPeriod = QueryPeriod(start, end)
        result = self.historyReadRaw(queryPeriod, True)

        previousVal = None
        exactVal = None
        nextVal = None

        if result.size == 1:
            if result[0][1] > timestamp:
                nextVal = result[0]
            elif result[0][1] == timestamp:
                exactVal = result[0]
            else:
                previousVal = result[0]
        elif result.size == 2:
            if result[1][1] > timestamp:
                nextVal = result[1]
            elif result[1][1] == timestamp:
                exactVal = result[1]
            if result[0][1] < timestamp:
                previousVal = result[0]
            elif result[0][1] == timestamp:
                exactVal = result[0]
        elif result.size == 3:
            previousVal = result[0]
            exactVal = result[1]
            nextVal = result[2]
        elif result.size == 4:
            previousVal = result[1]
            exactVal = result[2]
            nextVal = result[3]
        else:
            return None

        if retrieval == RetrievalMode.Previous:
            if previousVal and previousVal[2] == 2156527616:
                return self.recordedValue(previousVal[1], retrieval)
            else:
                return previousVal
        elif retrieval == RetrievalMode.Exact:
            if exactVal and exactVal[2] == 2156527616:
                return None
            else:
                return exactVal
        elif retrieval == RetrievalMode.Next:
            if nextVal and nextVal[2] == 2156527616:
                return self.recordedValue(nextVal[1], retrieval)
            else:
                return nextVal

    def readAnnotations(self, start, end):
      queryPeriod = QueryPeriod(start, end)
      annotationPath = ItemPathJSON('OPCUA.NodeId', None, self._encodePropertyIdentity(EpmDataObjectAttributeIds.Annotations.value))
      return self._epmConnection._historyReadAnnotation(queryPeriod, annotationPath)

    def deleteAnnotations(self, start, end, allUsers = False):
      userName = self._epmConnection._authorizationService._userName
      result = self.readAnnotations(start, end)
      deletedItems = []
      for index in range(len(result)):
        if allUsers or result[index][1] == userName:
          deletedItems.append(result[index])

      if len(deletedItems) == 0:
        return

      annotationPath = ItemPathJSON('OPCUA.NodeId', None, self._encodePropertyIdentity(EpmDataObjectAttributeIds.Annotations.value))
      from .performupdatetype import PerformUpdateType
      self._epmConnection._historyUpdateAnnotation(annotationPath, PerformUpdateType.Remove.value, deletedItems)

    def writeAnnotation(self, timestamp, message, override = True):
      timestamp = timestamp.astimezone(pytz.UTC)
      annotationPath = ItemPathJSON('OPCUA.NodeId', None, self._encodePropertyIdentity(EpmDataObjectAttributeIds.Annotations.value))
      userName = self._epmConnection._authorizationService._userName
      annotations = [ (timestamp, userName, message) ]
      from .performupdatetype import PerformUpdateType
      if override:
        results = self.readAnnotations(timestamp, timestamp + dt.timedelta(seconds=1))
        for index in range(len(results)):
          if results[index][1] == userName and self._compareDatetime(results[index][0].astimezone(pytz.UTC), timestamp):
            return self._epmConnection._historyUpdateAnnotation(annotationPath, PerformUpdateType.Update.value, annotations)
      return self._epmConnection._historyUpdateAnnotation(annotationPath, PerformUpdateType.Insert.value, annotations)

    def readAttributes(self):
      self._epmConnection._fillDataObjectsAttributes([ self ], DataObjectAttributes.All)

    def enumProperties(self):
      result = self._epmConnection._browse([ self._itemPath ], EpmNodeIds.HasProperty.value).references()
      childProperties = collections.OrderedDict()
      for item in result[0]:
        childProperties[item._displayName] = EpmProperty(self._epmConnection, item._displayName, self._path + '/' + item._displayName, item._identity)
      return childProperties

    def _compareDatetime(self, datetime1, datetime2):
      return (datetime1.year == datetime2.year and
              datetime1.month == datetime2.month and
              datetime1.day == datetime2.day and
              datetime1.hour == datetime2.hour and
              datetime1.minute == datetime2.minute and
              datetime1.second == datetime2.second and
              datetime1.microsecond == datetime2.microsecond)

    def _setAttribute(self, attributeId, value):
      if attributeId == DataObjectAttributes.Description:
        self._description = value
      elif attributeId == DataObjectAttributes.EU:
        self._eu = value['displayName']
      elif attributeId == DataObjectAttributes.HighLimit:
        self._highLimit = value
      elif attributeId == DataObjectAttributes.LowLimit:
        self._lowLimit = value
      elif attributeId == DataObjectAttributes.Clamping:
        self._clamping = value
      elif attributeId == DataObjectAttributes.Domain:
        self._domain = 'Discrete' if value else 'Continuous'


    #Private Methods
    def _encodePropertyIdentity(self, propertyIdentity):

      nodeIdSplitted = self._identity.split(';')
      if len(nodeIdSplitted) != 2:
        raise Exception('Invalid nodeId format')
      matches = [x for x in nodeIdSplitted if x.startswith('i=')]
      if len(matches) != 1:
        raise Exception('Invalid nodeId type')

      objectIdentity = int(matches[0][2:])

      SignatureLSB = 0xFE;
      SignatureMSB = 0xCA;

      ret = [ 0 ] * 8

      ret[0] = ((propertyIdentity & 0x000000FF) >> (0 * 8));
      ret[1] = ((propertyIdentity & 0x0000FF00) >> (1 * 8));
      ret[2] = ((objectIdentity   & 0x000000FF) >> (0 * 8));
      ret[3] = ((objectIdentity   & 0x0000FF00) >> (1 * 8));
      ret[4] = ((objectIdentity   & 0x00FF0000) >> (2 * 8));
      ret[5] = ((objectIdentity   & 0xFF000000) >> (3 * 8));
      ret[6] = SignatureLSB;
      ret[7] = SignatureMSB;

      import base64

      return 'ns=1;b=' + base64.b64encode(bytes(ret)).decode("utf-8") 







"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import datetime as dt
import requests
import json
from .autostarttimer import AutoStartTimer

class AuthorizationService(object):

  def __init__(self, authServer, clientId, programId, userName, password):
    self._authServer = authServer
    self._clientId = clientId
    self._programId = programId
    self._userName = userName
    self._password = password
    self._refreshToken = None
    self._tokenExpiration = None
    self._token = None
    self._token = self.renewToken()
    self._timer = AutoStartTimer(180, self.reloadToken)
    self._error = ''

  def reloadToken(self):
    try:
      self._token = self.renewToken()
      self._error = ''
    except Exception as ex:
      self._token = ''
      self._error = str(ex)
    
  def renewToken(self):
    if (self._token != None):
      return self.refreshToken()

    client_auth = requests.auth.HTTPBasicAuth(self._clientId, self._programId)
    post_data = {"grant_type": "password", 
                  "username": self._userName,
                  "password": self._password,
                  "scope": "offline_access openid profile email opcua_browse opcua_read opcua_write opcua_subscription opcua_history EpmWebApi portal_files portal_upload"} #EpmProcessor #openid profile email opcua_browse opcua_read opcua_subscription 
    auth_url = self._authServer + '/connect/token'
    response = requests.post(auth_url,
                             auth=client_auth,
                             data=post_data, verify=False)
    respose_json = response.json()

    if response.status_code != 200:
        raise Exception("GetToken() call http error '" +  str(response.status_code) + "'. Reason: " + respose_json["error"])

    self._refreshToken = respose_json["refresh_token"]

    return respose_json["access_token"]

  def getToken(self):
    if self._error == '':
      return self._token
    else:
      raise Exception(self._error)

  def refreshToken(self):
    client_auth = requests.auth.HTTPBasicAuth(self._clientId, self._programId)
    post_data = {"grant_type": "refresh_token", 
                 "refresh_token": "%s"%(self._refreshToken) }
    auth_url = self._authServer + '/connect/token'
    response = requests.post(auth_url,
                              auth=client_auth,
                              data=post_data, verify=False) 
    respose_json = response.json()

    if response.status_code != 200:
        raise Exception("RefreshToken() call http error '" +  str(response.status_code) + "'. Reason: " + respose_json["error"])

    
    self._refreshToken = respose_json["refresh_token"]
    return respose_json["access_token"]

  def logout(self):
    if (self._token == None):
      return
    post_data = { "token" : self._token }
    client_auth = requests.auth.HTTPBasicAuth(self._clientId, self._programId)
    auth_url = self._authServer + '/connect/revocation'
    response = requests.post(auth_url,
                              auth=client_auth,
                              data=post_data, verify=False)
    if (self._refreshToken == None):
      return
    post_data = { "token" : self._refreshToken, "token_type_hint" : "refresh_token" }
    client_auth = requests.auth.HTTPBasicAuth(self._clientId, self._programId)
    auth_url = self._authServer + '/connect/revocation'
    response = requests.post(auth_url,
                              auth=client_auth,
                              data=post_data, verify=False)

  def close(self):
    try:
      self._timer.cancel()
    except Exception as ex:
      import logging
      logging.exception(ex)
    self.logout()

  def restart(self):
    self._refreshToken = None
    self._tokenExpiration = None
    self._token = None
    self._token = self.renewToken()
    self._timer = AutoStartTimer(60, self.reloadToken)






"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class ReadModelJSON(object):

  def __init__(self, items, continuationPoint):
    self._items = items
    self._continuationPoint = continuationPoint

  def toDict(self):

    childMap = []
    for item in self._items:
        childMap.append(item.toDict());

    return {'items': childMap, 'continuationPoint': self._continuationPoint }



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .domainfilter import DomainFilter

from enum import IntFlag


class DataObjectsFilterType(IntFlag):

  BasicVariable = 1

  ExpressionVariable = 2


class DataObjectsFilter(object):
  """description of class"""

  def __init__(self, type = None, name = None, eu = None, description = None, domain = None):
    self._type = DataObjectsFilterType.BasicVariable | DataObjectsFilterType.ExpressionVariable if type == None else type
    self._name = '*' if name == None else name
    self._eu = '*' if eu == None else eu
    self._description = description
    self._domain = DomainFilter.All if domain == None else domain

  @property
  def type(self):
    return self._type

  @property
  def name(self):
    return self._name
  
  @property
  def eu(self):
    return self._eu

  @property
  def description(self):
    return self._description

  @property
  def domain(self):
    return self._domain



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import json
class QueryResultModelJSON(object):
    """description of class"""
    def __init__(self, items, diagnostic, continuationPoint):
      self._items = items
      self._diagnostic = diagnostic
      self._continuationPoint = continuationPoint

    def toDict(self):
        jsonPaths = []

        items = []
        for item in self._items:
            items.append(item.toDict());

        return {'items' : items, 'diagnostic' : self._diagnostic.toDict(), 'continuationPoint': self._continuationPoint }




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class HistoryUpdateDataItemModelJSON(object):
    """description of class"""
    def __init__(self, path, updateType, values, typePath = None):
        self._path = path
        self._typePath = typePath
        self._updateType = updateType
        self._values = values

    def toDict(self):
        jsonValues = []
        for value in self._values:
            jsonValues.append(value.toDict());

        typePath = None
        if self._typePath != None:
          typePath = self._typePath.toDict()

        return {'path': self._path.toDict(), 'typePath' : typePath, 'updateType': self._updateType, 
               'values' : jsonValues }




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class HistoryUpdateDataModelJSON(object):
    """description of class"""
    def __init__(self, items):
        self._items = items

    def toDict(self):
        jsonItems = []

        for item in self._items:
            jsonItems.append(item.toDict());

        return {'items': jsonItems }





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .itempathjson import ItemPathJSON
import datetime as dt
from .datavaluejson import DataValueJSON
from .historyupdatetype import HistoryUpdateType

class EpmVariable(object):
    """description of class"""

    def __init__(self, epmConnection, name, path, identity):
      self._epmConnection = epmConnection
      self._name = name
      self._path = path
      self._itemPath = ItemPathJSON('OPCUA.NodeId', None, identity)

    ### Properties
    @property
    def name(self):
        return self._name

    @property
    def path(self, value):
        self._path = value

    ### Methods

    def historyReadRaw(self, queryPeriod, bounds = False):
        return self._epmConnection._historyReadRaw(queryPeriod, self._itemPath, bounds = bounds)

    def read(self):
        readResult = self._epmConnection._read([self._itemPath], [13]).items()[0]
        if readResult[1].code != 0:
          raise Exception("Read from '" + self._path + "' failed with error: " + str(readResult[1].code))
        return readResult[0].value

    def write(self, value, timestamp=dt.datetime.now(dt.timezone.utc), quality=0):
        return self._epmConnection._write([self._itemPath], [13], [DataValueJSON(value, quality, timestamp)])

    def historyReadAggregate(self, aggregateDetails, queryPeriod):
        return self._epmConnection._historyReadAggregate(aggregateDetails, queryPeriod, self._itemPath)

    def historyUpdate(self, values):
        return self._epmConnection._historyUpdate(HistoryUpdateType.Update.value, [ self._itemPath ], [ values ])

    def historyDelete(self, queryPeriod):
        values = self._epmConnection._historyReadRaw(queryPeriod, self._itemPath, False)
        return self._epmConnection._historyUpdate(HistoryUpdateType.Remove.value, [ self._itemPath ], [ values ])





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import requests
import json

class Resource(object):

  def __init__(self, resourcesManager, id, name, description, mimeType):
    self._resourcesManager = resourcesManager
    self._id = id
    self._name = name
    self._description = description
    self._mimeType = mimeType

  def download(self, type):

    return self._resourcesManager._download(self._id, type)

  def upload(self, file):

    return self._resourcesManager._updateFile(self._id, self._name, self._mimeType, file)

  def delete(self):

    return self._resourcesManager._delete(self._id)





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class BrowseResultItemModelJSON(object):

  def __init__(self, identity, displayName, relativePath, type, nodeClass):
    self._identity = identity
    self._displayName = displayName
    self._relativePath = relativePath
    self._type = type
    self._nodeClass = nodeClass




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class ReadResultItemModelJSON(object):

  def __init__(self, identity, value):
    self._identity = identity
    self._value = value

  @property
  def identity(self):
      return self._identity

  @property
  def value(self):
      return self._value




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class NodeAttributes(Enum):

    # <summary>
    # The canonical identifier for the node.
    # </summary>
    NodeId = 1

    # <summary>
    # The class of the node.
    # </summary>
    NodeClass = 2

    # <summary>
    # A non-localized, human readable name for the node.
    # </summary>
    BrowseName = 3

    # <summary>
    # A localized, human readable name for the node.
    # </summary>
    DisplayName = 4

    # <summary>
    # A localized description for the node.
    # </summary>
    Description = 5

    # <summary>
    # Indicates which attributes are writeable.
    # </summary>
    WriteMask = 6

    # <summary>
    # Indicates which attributes are writeable by the current user.
    # </summary>
    UserWriteMask = 7

    # <summary>
    # Indicates that a type node may not be instantiated.
    # </summary>
    IsAbstract = 8

    # <summary>
    # Indicates that forward and inverse references have the same meaning.
    # </summary>
    Symmetric = 9

    # <summary>
    # The browse name for an inverse reference.
    # </summary>
    InverseName = 10

    # <summary>
    # Indicates that following forward references within a view will not cause a loop.
    # </summary>
    ContainsNoLoops = 11

    # <summary>
    # Indicates that the node can be used to subscribe to events.
    # </summary>
    EventNotifier = 12

    # <summary>
    # The value of a variable.
    # </summary>
    Value = 13

    # <summary>
    # The node id of the data type for the variable value.
    # </summary>
    DataType = 14

    # <summary>
    # The number of dimensions in the value.
    # </summary>
    ValueRank = 15

    # <summary>
    # The length for each dimension of an array value.
    # </summary>
    ArrayDimensions = 16

    # <summary>
    # How a variable may be accessed.
    # </summary>
    AccessLevel = 17

    # <summary>
    # How a variable may be accessed after taking the user's access rights into account.
    # </summary>
    UserAccessLevel = 18

    # <summary>
    # Specifies (in milliseconds) how fast the server can reasonably sample the value for changes.
    # </summary>
    MinimumSamplingInterval = 19

    # <summary>
    # Specifies whether the server is actively collecting historical data for the variable.
    # </summary>
    Historizing = 20

    # <summary>
    # Whether the method can be called.
    # </summary>
    Executable = 21

    # <summary>
    # Whether the method can be called by the current user.
    # </summary>
    UserExecutable = 22



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class PerformUpdateType(Enum):

    Insert = 0

    Remove = 1

    Replace = 2

    Update = 3




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class ItemPathAndContinuationPointJSON(object):
    """description of class"""
    def __init__(self, path, continuationPoint, release):
        self._path = path
        self._continuationPoint = continuationPoint
        self._release = release

    def toDict(self):
        return {'path' : self._path.toDict(), 'continuationPoint' : self._continuationPoint, 'release' : self._release}




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import dateutil
import datetime
from datetime import timezone

class DataValueJSON(object):
    """description of class"""
    def __init__(self, value, statusCode, timestamp):
        self._value = value
        self._statusCode = statusCode
        if type(timestamp) == str:
          try:
            self._timestamp = dateutil.parser.parse(timestamp).astimezone(timezone.utc)
          except OverflowError as error:
            self._timestamp = datetime.datetime(1,1,1,0,0,tzinfo=datetime.timezone.utc)
        else:
          self._timestamp = timestamp

    @property
    def value(self):
      return self._value

    @property
    def statusCode(self):
      return self._statusCode

    @property
    def timestamp(self):
      return self._timestamp

    def toDict(self):
        return {'value': self._value, 'quality': self._statusCode, 
               'timestamp' : self._timestamp.isoformat() }



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class Application(Enum):
    Atom =                                "application/atom+xml";
    Compress =                            "application/x-compress";
    Compressed =                          "application/x-compressed";
    DirectX =                             "application/directx";
    ElipsePortalDashboard =               "application/vnd.elipse.portal.dashboard";
    ElipsePortalDashboardThumbnail =      "application/vnd.elipse.portal.dashboard.thumbnail";
    ElipsePortalFolder =                  "application/vnd.elipse.portal.folder";
    Gtar =                                "application/x-gtar";
    Gzip =                                "application/x-gzip";
    InternetStream =                      "application/internet-property-stream";
    JavaArchive =                         "application/java-archive";
    Javascript =                          "application/x-javascript";
    Json =                                "application/json";
    MicrosoftAccess =                     "application/msaccess";
    MicrosoftAccess2 =                    "application/x-msaccess";
    MicrosoftExcel =                      "application/vnd.ms-excel";
    MicrosoftPowerpoint =                 "application/vnd.ms-powerpoint";
    MicrosoftWord =                       "application/msword";
    OctetStream =                         "application/octet-stream";
    OneNote =                             "application/onenote";
    Pdf =                                 "application/pdf";
    Postscript =                          "application/postscript";
    PerformanceMonitor =                  "application/x-perfmon";
    Rtf =                                 "application/rtf";
    ShockwaveFlash =                      "application/x-shockwave-flash";
    Silverlight =                         "application/x-silverlight-app";
    Tar =                                 "application/x-tar";
    Visio =                               "application/vnd.visio";
    Xaml =                                "application/xaml+xml";
    Zip =                                 "application/x-zip-compressed";

class Text(Enum):
    Css =                                 "text/css";
    Html =                                "text/html";
    JScript =                             "text/jscript";
    Plain =                               "text/plain";
    RichText =                            "text/richtext";
    Scriptlet =                           "text/scriptlet";
    TabSeparatedValues =                  "text/tab-separated-values";
    VbScript =                            "text/vbscript";
    Xml =                                 "text/xml";

class Image(Enum):
    Bmp =                                 "image/bmp";
    Gif =                                 "image/gif";
    Icon =                                "image/x-icon";
    Ief =                                 "image/ief";
    Jpeg =                                "image/jpeg";
    Pjpeg =                               "image/pjpeg";
    Png =                                 "image/png";
    Tiff =                                "image/tiff";

class Audio(Enum):
    Aiff =                                "audio/aiff";
    Basic =                               "audio/basic";
    Mid =                                 "audio/mid";
    Mpeg =                                "audio/mpeg";
    Wav =                                 "audio/wav";

class Video(Enum):
    Mpeg =                                "video/mpeg";
    Quicktime =                           "video/quicktime";






"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .resourcesmanager import ResourcesManager

class PortalResources(ResourcesManager):

  def __init__(self, authorizationService, webApi):
    super().__init__(authorizationService, webApi, 'resources')



    




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import json
class HistoryRawModelJSON(object):
    """description of class"""
    def __init__(self, startTime, endTime, bounds, paths):
        self._startTime = startTime
        self._endTime = endTime
        self._bounds = bounds
        self._paths = paths

    def toDict(self):
        jsonPaths = []

        childMap = []
        for item in self._paths:
            childMap.append(item.toDict());

        map = {'startTime' : self._startTime.isoformat(), 'endTime' : self._endTime.isoformat(), 
                                'bounds' : self._bounds, 'paths' : childMap}

        return map




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class BrowseResultModelJSON(object):
  
  def __init__(self, items, diagnostics):
    self._items = items
    self._diagnostics = diagnostics

  def items(self):
    return list(zip(self._items, self._diagnostics))

  def diagnostics(self):
    return self._diagnostics

  def references(self):
    return self._items




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .itempathjson import ItemPathJSON
import datetime as dt
from .datavaluejson import DataValueJSON
from .epmvariable import EpmVariable

class EpmProperty(EpmVariable):
    """description of class"""

    def __init__(self, epmConnection, name, path, identity):
        super().__init__(epmConnection, name, path, identity)









"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import datetime
from enum import Enum
from dateutil.relativedelta import relativedelta
import re
import os

from .datasetpen import DatasetPen
from .queryperiod import QueryPeriod

class PeriodUnit(Enum):
    Second = 1

    Minute = 2

    Hour = 3

    Day = 4

    Month = 5


class DatasetConfig(object):
    """description of class"""
    def __init__(self, connection, name, content = None, description = None, filePath = None):
        import clr
        import pytz
        clr.AddReference('dll_references/EpmData')
        from Elipse.Epm.Common import EpmExtensionObject, SerializationExtensions, TimeInterval, DatasetAndChartData, \
                                      DatasetData
        self._connection = connection
        self._name = None
        self._count = None
        self._periodUnit = None
        self._startTime = None
        self._endTime = None
        self._period = None
        self._datasetPens = []
        self._filePath = filePath
        self._nameChanged = False
        if content is None:
            self.setName(name)
            self.recentPeriodConfig(1, PeriodUnit.Hour)
            self._description = description
            self._datasetAndChartData = DatasetAndChartData(DatasetData(), '')
        else:
            self._name = name
            contentObject = EpmExtensionObject.FromXml(content)
            resource = SerializationExtensions.ReadFileResourceContent(contentObject)
            self._description = resource.Description
            resourceObject = EpmExtensionObject.FromXml(resource.Content)
            dataset = SerializationExtensions.ReadDatasetAndChartData(resourceObject)
            self._datasetAndChartData = dataset

            periodType = dataset.Dataset.Period
            if isinstance(periodType, TimeInterval):
                self._isTimeInterval = True
                start = dataset.Dataset.Period.StartTime
                startUtc = start.ToUniversalTime()
                startTime = datetime.datetime.strptime(startUtc.ToString(), '%m/%d/%Y %H:%M:%S %p')
                startTime = pytz.UTC.localize(startTime)
                self._startTime = startTime
                end = dataset.Dataset.Period.EndTime
                endUtc = end.ToUniversalTime()
                endTime = datetime.datetime.strptime(endUtc.ToString(), '%m/%d/%Y %H:%M:%S %p')
                endTime = pytz.UTC.localize(endTime)
                self._endTime = endTime
            else:
                self._isTimeInterval = False
                count = dataset.Dataset.Period.Count
                period = dataset.Dataset.Period.PeriodType
                self._periodXmlToTimePeriod(count, period)


            for penConfig in dataset.Dataset.Fields:
                title = penConfig.Alias
                self._datasetPens.append(DatasetPen(self, title, penConfig=penConfig))


    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def startTime(self):
      return self._startTime

    @property
    def endTime(self):
      return self._endTime

    @property
    def period(self):
        return self._period

    @property
    def isTimeInterval(self):
        return self._isTimeInterval

    @property
    def datasetPens(self):
      return self._datasetPens

    @property
    def filePath(self):
        return self._filePath

    def setName(self, name):
        if type(name) is str:
            pattern = r'^[:><"/\*]+$'
            if re.search(pattern, name):
                raise Exception("Invalid character on string argument")
            else:
                if name != self.name:
                    self._name = name
                    self._nameChanged = True
        else:
            raise Exception("Argument must be a string")

    def setDescription(self, description):
        if type(description) is str:
            if len(description) > 1024:
                raise Exception("Description cannot exceed 1024 characters")
            else:
                self._description = description
        else:
            raise Exception("Argument must be a string")

    def timeIntervalConfig(self, startTime, endTime):
        if isinstance(startTime, datetime.datetime) and isinstance(endTime, datetime.datetime):
            import pytz
            startTimeUtc = startTime.astimezone(pytz.UTC)
            endTimeUtc = endTime.astimezone(pytz.UTC)
            if startTimeUtc < endTimeUtc:
                self._startTime = startTimeUtc
                self._endTime = endTimeUtc
                self._isTimeInterval = True
            else:
                raise Exception("startTime must be before endTime")
        else:
            raise Exception("Arguments must be of datetime type")

    def recentPeriodConfig(self, count, periodUnit):
        if not isinstance(count, int):
            raise Exception("count argument must be an integer")
        if not isinstance(periodUnit, PeriodUnit):
            raise Exception("periodUnit argument must be of type PeriodUnit")
        self._periodXmlToTimePeriod(count, periodUnit.value)
        self._isTimeInterval = False

    def addPen(self, title, dataSourceName = None):
        for pen in self._datasetPens:
            if pen.title == title:
                raise Exception("Pen title already exists")
        datasetPen = DatasetPen(self, title, dataSourceName)
        self._datasetPens.append(datasetPen)
        return self.getPen(title)

    def getPen(self, title):
        for pen in self._datasetPens:
            if pen.title == title:
                return pen
        raise Exception("Pen not found")

    def removePen(self, title):
        for pen in self._datasetPens:
            if pen.title == title:
                self._datasetPens.remove(pen)
                return
        raise Exception("Pen not found")

    def execute(self):
        results = {}

        if self.isTimeInterval:
            queryPeriod = QueryPeriod(self.startTime, self.endTime)
        else:
            import pytz
            endTime = datetime.datetime.utcnow()
            endTime = pytz.UTC.localize(endTime)
            startTime = endTime - self.period
            queryPeriod = QueryPeriod(startTime, endTime)

        for pen in self.datasetPens:
            pen.setDataSource(pen.dataSource.name)
            if pen.isRaw:
                result = self._connection._historyReadRaw(queryPeriod, pen.dataSource._itemPath)
            else:
                result = self._connection._historyReadAggregate(pen.aggregateType, queryPeriod, pen.dataSource._itemPath)
            results[pen.title] = result

        return results

    def save(self):
        epmExtensionXml = self._getEpmExtensionXml()
        if self._filePath is not None:
            if self._nameChanged:
                pathSplit = self._filePath.split("/")
                pathSplit.pop(-1)
                path = '/'.join(pathSplit)
                newFilePath = path + '/' + self.name + ".epmdataset"
                if os.path.exists(newFilePath):
                    raise Exception("Dataset name changed but there is another dataset with this name on the same folder")
                self._connection._saveDatasetFile(epmExtensionXml, self._filePath, overwrite=True)
                os.rename(self._filePath, newFilePath)
                self._filePath = newFilePath
                self._nameChanged = False
            else:
                self._connection._saveDatasetFile(epmExtensionXml, self._filePath, overwrite=True)
        else:
            raise Exception("There is no dataset file to save, use the method saveToFile() instead")

    def saveToFile(self, path, overwrite = False):
        if path.endswith('/'):
            filePath = path + self.name + ".epmdataset"
        else:
            filePath = path + '/' + self.name + ".epmdataset"
        epmExtensionXml = self._getEpmExtensionXml()
        self._connection._saveDatasetFile(epmExtensionXml, filePath, overwrite=overwrite)
        self._filePath = filePath

    def delete(self):
        self._connection._deleteDatasetFile(self._filePath)
        self._filePath = None

    def duplicate(self, newName, samePath = False):
        duplicate = DatasetConfig(self._connection, self._name, self._getEpmExtensionXml(), filePath = self._filePath)
        duplicate.setName(newName)
        if samePath:
            if self._filePath is None:
                raise Exception("Dataset has no file on folder")
            pathSplit = self._filePath.split("/")
            pathSplit.pop(-1)
            path = '/'.join(pathSplit)
            newFilePath = path + '/' + newName + ".epmdataset"
            if os.path.exists(newFilePath):
                raise Exception("Dataset name already exists on that folder")
            duplicate._filePath = newFilePath
        else:
            duplicate._filePath = None

        return duplicate

    def _getEpmExtensionXml(self):
        import clr
        clr.AddReference('dll_references/EpmData')
        from Elipse.Epm.Common import TimeInterval, RecentPeriod, DatasetField, RawByPeriod, TrendMode, CalculateMode, \
            EpmExtensionObject, DatasetAndChartData, FileContentData, DatasetData
        from Elipse.Epm.AddressSpaceModel import NodeIdentifier
        from System import DateTime, TimeSpan
        from System.Collections.Generic import List

        if self.isTimeInterval:
            period = TimeInterval()
            period.StartTime = DateTime.Parse(str(self.startTime))
            period.EndTime = DateTime.Parse(str(self.endTime))
        else:
            period = RecentPeriod()
            period.Count = self._count
            period.PeriodType = self._periodUnit

        fields = []
        for datasetPen in self._datasetPens:
            pen = DatasetField()
            pen.Alias = datasetPen.title
            if datasetPen.isRaw:
                pen.Mode = RawByPeriod()
            else:
                sampleInterval = TimeSpan.Parse(str(datasetPen.aggregateType.interval))
                aggregateType = datasetPen._getAggregateId()
                if aggregateType == 1:
                    pen.Mode = TrendMode(sampleInterval)
                else:
                    pen.Mode = CalculateMode(aggregateType, sampleInterval)
            nodeId = datasetPen.dataSource._identity
            identifier = int(nodeId.split(';')[1].split('=')[1])
            namespaceIndex = int(nodeId.split(';')[0].split('=')[1])
            pen.Identity = NodeIdentifier()
            pen.Identity.IdentifierType = 1
            pen.Identity.Numeric32Identifier = identifier
            pen.Identity.NamespaceIndex = namespaceIndex
            fields.append(pen)

        self._datasetAndChartData.Dataset.Period = period
        fieldsList = List[DatasetField]()
        for field in fields:
            fieldsList.Add(field)
        self._datasetAndChartData.Dataset.Fields = fieldsList

        epmExtensionObject = EpmExtensionObject(self._datasetAndChartData.ToXml())
        epmExtensionObject.Tag = DatasetAndChartData.Tag
        epmExtensionXml = epmExtensionObject.ToXml()

        fileContent = FileContentData(self.description, epmExtensionXml)
        fileContentXml = fileContent.ToXml()

        epmExtensionObject = EpmExtensionObject()
        epmExtensionObject.Tag = FileContentData.Tag
        epmExtensionObject.Content = fileContentXml
        epmExtensionXml = epmExtensionObject.ToXml()

        return epmExtensionXml

    def _periodXmlToTimePeriod(self, count, period):
        if period == 1:
            self._period = datetime.timedelta(seconds=count)
        elif period == 2:
            self._period = datetime.timedelta(minutes=count)
        elif period == 3:
            self._period = datetime.timedelta(hours=count)
        elif period == 4:
            self._period = datetime.timedelta(days=count)
        elif period == 5:
            self._period = relativedelta(months=count)
        else:
            raise Exception("Error with the period time unit")
        self._count = count
        self._periodUnit = period



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import requests
import json

class Folder(object):

  def __init__(self, resourcesManager, id, name, description):
    self._resourcesManager = resourcesManager
    self._id = id
    self._name = name
    self._description = description


  def browse(self):

    return self._resourcesManager._browse(self._id)

  def upload(self, name, file, description = None, mimeType = None, thumbnail = None, thumbnailMimeType = None, starred = None, overrideFile=False):
    
    return self._resourcesManager._upload(self._id, name, file, description, mimeType, thumbnail, thumbnailMimeType, starred, overrideFile)

  def createFolder(self, name, description):
    
    return self._resourcesManager._createFolder(self._id, name, description)

  def delete(self):
    return self._resourcesManager._delete(self._id);



from enum import IntFlag

class BasicVariablePropertyMask(IntFlag):

    Unspecified = 0,

    Name = 1,

    Description = 1,

    TagType = 2,

    RealTimeEnabled = 4,

    DeadBandFilter = 8,

    DeadBandUnit = 16,

    Eu = 32,

    LowLimit = 64, 

    HighLimit = 128,

    ScaleEnable = 256,

    InputLowLimit = 512,

    InputHighLimit = 1024,

    Clamping = 2048,

    Domain = 4096,

    Interface = 8192,

    IoTagAddress = 16384,

    ProcessingEnabled = 32768,

    IsRecording = 65536,

    IsCompressing = 131072,

    StoreMillisecondsEnabled = 262144,

    StorageSet = 524288,







"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .epmdataobject import EpmDataObject
from .itempathjson import ItemPathJSON
from .epmproperty import EpmProperty
from .epmnodeids import EpmNodeIds

import collections


class EpmObject(object):
    """description of class"""

    def __init__(self, epmConnection, identity, path, name, type = ''):
        self._epmConnection = epmConnection
        self._itemPath = ItemPathJSON('OPCUA.NodeId', None, identity)
        self._path = path
        self._type = type
        self._name = name

    # Public Methods

    #def getProperty(self, property):
    #    if self._propertiesNames.index(property) < 0:
    #        raise Exception('Invalid property name')
    #    return EpmProperty(self._epmConnection, property, self._browsePath + '/1:' + property)

    def enumObjects(self):
        result = self._epmConnection._browse([self._itemPath], EpmNodeIds.HasComponent.value).references()[0]
        if len(result) < 1:
            return []
        childObjects = collections.OrderedDict()
        types = {}
        for item in result:
            if item._nodeClass == 4:  # Method is ignored
                continue
            childObjects[item._displayName] = EpmObject(self._epmConnection, item._identity,
                                                        self._path + '/' + item._displayName, item._displayName,
                                                        item._type)
            types[item._type] = ItemPathJSON('OPCUA.NodeId', '', item._type)
        typeNames = self._epmConnection._read(list(types.values()), [4] * len(types.values())).values()
        for x in typeNames:
            types[x.identity] = x.value.value
        for x in childObjects:
            childObjects[x].type = types[childObjects[x].type]

        return childObjects

    def enumProperties(self):
      result = self._epmConnection._browse([ self._itemPath ], EpmNodeIds.HasProperty.value)
      childProperties = collections.OrderedDict()
      for item in result.references()[0]:
        childProperties[item._displayName] = EpmProperty(self._epmConnection, item._displayName, self._path + '/' + item._displayName, item._identity)
      return childProperties

    # Public Properties

    @property
    def name(self):
        return self._name

    @property
    def path(self):
        return self._path

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import json
class ItemPathJSON(object):
    """description of class"""
    def __init__(self, language, context, path):
        self._language = language
        self._context = context
        self._relativePath = path

    @property
    def relativePath(self):
      return self._relativePath;

    def toDict(self):
        return {'language' : self._language, 'context' : self._context, 'path' : self._relativePath}

    def toJSON(self):
        return json.dumps({'language' : self._language, 'context' : self._context, 'path' : self._relativePath})




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import numpy as np
import datetime as dt
import dateutil.parser
import datetime
from datetime import timezone

class NumpyExtras:

    def __init__(self):
      self._int = {}
      for i in range(60):
        self._int['{:02}'.format(i)] = i

    def numpyArrayFromDataValues(self, dataValues):
        valuesCount = len(dataValues)
        i = 0

        dataType = float
        while (i < valuesCount):
            if (dataValues[i]['value'] != None):
                dataType = type(dataValues[i]['value'])
                break;
            i = i + 1

        numpyArray = self._getNumpyArray(dataType, valuesCount)
        if valuesCount < 1:
            return numpyArray

        import time
        start_time = time.time()

        timestamps = map(self.fastStrptime3, dataValues)
        numpyArray['Timestamp'].flat[:] = list(timestamps)

        numpyArray['Quality'].flat[:] = [ x['quality'] for x in dataValues]

        if dataType == float or dataType == int:
          numpyArray['Value'].flat[:] = list(map(lambda x: x['value'] if x['value'] != None else 0, dataValues))
        else:
          i = 0
          for numpyValue in numpyArray:
            dataValue = dataValues[i]
            if dataValue['value'] != None:
              numpyValue['Value'] = dataValue['value']
            i = i + 1

        return numpyArray

    def _getNumpyArray(self, dataType, valuesCount):
        if dataType == int:
            return np.empty([valuesCount], dtype=np.dtype([('Value', '>i8'), ('Timestamp', 'object'), ('Quality', '>i8')]))
        elif dataType == float:
            return np.empty([valuesCount], dtype=np.dtype([('Value', '>f4'), ('Timestamp', 'object'), ('Quality', '>i8')]))
        elif dataType == bool:
            return np.empty([valuesCount], dtype=np.dtype([('Value', 'bool'), ('Timestamp', 'object'), ('Quality', '>i8')]))
        elif dataType == str:
            return np.empty([valuesCount], dtype=np.dtype([('Value', 'object'), ('Timestamp', 'object'), ('Quality', '>i8')]))
        else: 
            return np.empty([valuesCount], dtype=np.dtype([('Value', '>f8'), ('Timestamp', 'object'), ('Quality', '>i8')]))

    def fastStrptime3(self, item) -> datetime.datetime:
      val = item['timestamp']
      l = len(val)
      if (l == 23 or l == 24): #format == "%Y-%m-%dT%H:%M:%S.%fZ" and 
          us = int(val[20:(l - 1)])
          # If only milliseconds are given we need to convert to microseconds.
          if l == 23:
              us *= 10000
          if l == 24:
              us *= 1000
          return datetime.datetime(*map(int, [val[0:4], val[5:7], val[8:10], val[11:13], val[14:16], val[17:19]]), us, tzinfo=timezone.utc
          )
      elif (l == 20): #format == "%Y-%m-%dT%H:%M:%SZ" and 
          return datetime.datetime(*map(int, [val[0:4], val[5:7], val[8:10], val[11:13], val[14:16], val[17:19]]), 0, tzinfo=timezone.utc)
      else:
        return dateutil.parser.parse(val).astimezone(timezone.utc)

    def fastStrptime2(self, item) -> datetime.datetime:
      val = item['timestamp']
      l = len(val)
      if (l == 23 or l == 24): #format == "%Y-%m-%dT%H:%M:%S.%fZ" and 
          us = int(val[20:(l - 1)])
          # If only milliseconds are given we need to convert to microseconds.
          if l == 23:
              us *= 10000
          if l == 24:
              us *= 1000
          return datetime.datetime(
              int(val[0:4]), # %Y
              self._int[val[5:7]], # %m
              self._int[val[8:10]], # %d
              self._int[val[11:13]], # %H
              self._int[val[14:16]], # %M
              self._int[val[17:19]], # %s
              us, # %f
              tzinfo=timezone.utc
          )
      elif (l == 20): #format == "%Y-%m-%dT%H:%M:%SZ" and 
          return datetime.datetime(
              int(val[0:4]), # %Y
              self._int[val[5:7]], # %m
              self._int[val[8:10]], # %d
              self._int[val[11:13]], # %H
              self._int[val[14:16]], # %M
              self._int[val[17:19]], # %s
              0, # %f
              tzinfo=timezone.utc
          )
      else:
        return dateutil.parser.parse(str)

    def fastStrptime(self, val: str) -> datetime.datetime:
      l = len(val)
      if (l == 23 or l == 24): #format == "%Y-%m-%dT%H:%M:%S.%fZ" and 
          us = int(val[20:(l - 1)])
          # If only milliseconds are given we need to convert to microseconds.
          if l == 23:
              us *= 10000
          if l == 24:
              us *= 1000
          return datetime.datetime(
              int(val[0:4], 10), # %Y
              int(val[5:7], 10), # %m
              int(val[8:10], 10), # %d
              int(val[11:13], 10), # %H
              int(val[14:16], 10), # %M
              int(val[17:19], 10), # %s
              us, # %f
              tzinfo=timezone.utc
          )
      elif (l == 20): #format == "%Y-%m-%dT%H:%M:%SZ" and 
          return datetime.datetime(
              int(val[0:4], 10), # %Y
              int(val[5:7], 10), # %m
              int(val[8:10], 10), # %d
              int(val[11:13], 10), # %H
              int(val[14:16], 10), # %M
              int(val[17:19], 10), # %s
              0, # %f
              tzinfo=timezone.utc
          )
      else:
        return dateutil.parser.parse(str)

    def getDictValue(self, value):
      if value in self._int:
        return self._int[value]
      value = int(value)
      self._int[value] = value
      return value




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import json
class QueryResultItemModelJSON(object):
    """description of class"""
    def __init__(self, identity, name, description, typeId):
      self._identity = identity
      self._name = name
      self._typeId = typeId





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from ._version import __version__

from .epmconnection import EpmConnection, ImportMode, BadArgumentException
from .epmdataobject import EpmDataObject, RetrievalMode
from .epmmodelobject import EpmModelObject, InstanceNameDuplicatedException, InvalidObjectNameException, \
    InvalidObjectPropertyException, InvalidObjectTypeException, InvalidSourceVariableException, \
    ObjectDependenciesException
from .basicvariable import BasicVariable
from .epmobject import EpmObject
from .queryperiod import QueryPeriod
from .aggregatedetails import AggregateDetails
from .aggregatedetails import AggregateType
from .browseitemmodeljson import BrowseItemModelJSON
from .browsemodeljson import BrowseModelJSON
from .readitemmodeljson import ReadItemModelJSON
from .readmodeljson import ReadModelJSON
from .writeitemmodeljson import WriteItemModelJSON
from .writemodeljson import WriteModelJSON
from .querymodeljson import QueryModelJSON
from .domainfilter import DomainFilter
from .downloadtype import DownloadType
from .customtypedefinition import CustomTypeDefinition, CustomTypeAlreadyExistsException, \
    CustomTypeDependenciesException, DuplicatedPropertiesNamesException, DuplicatedPropertiesTypeException, \
    InvalidCustomTypeNameException, InvalidIconException, InvalidPropertyNameException, InvalidPropertyTypeException, \
    MissingPropertyNameException
from .datasetconfig import DatasetConfig, PeriodUnit
from .datasetpen import DatasetPen



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import Enum

class EpmNodeIds(Enum):

    HasTags = 'ns=4;i=96'

    Organizes = 'ns=0;i=35'

    HasComponent = 'ns=0;i=47'

    HasProperty = 'ns=0;i=46'

    BasicVariableType = 'ns=4;i=94'

    ExpressionVariableType = 'ns=4;i=289'

    AnnotationType = 'ns=0;i=891'

    HasTypeDefinition = 'ns=0;i=40'





from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name = 'epmwebapi', 
      version = '1.5.43', 
      author = 'Elipse Software Ltda', 
      packages = ['epmwebapi'], 
      include_package_data=True,
      description = '', 
      long_description = long_description, 
      long_description_content_type='text/markdown',
      url = 'https://github.com/elipsesoftware/epmwebapi',
      python_requires='>=3.6, <4',
      install_requires=[
          'numpy',
          'requests',
          'python-dateutil',
          'pytz',
          'pythonnet',
      ],
      data_files = [("", ["LICENSE.txt"]),("", ["Readme.md"]),("dll_references", ["epmwebapi/dll_references/Elipse.Epm.AddressSpaceModel.dll"]),("dll_references", ["epmwebapi/dll_references/EpmData.dll"])],
      classifiers=[
      'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Intended Audience :: Developers',      # Define that your audience are developers
      'Topic :: Software Development :: Build Tools',
      'License :: OSI Approved :: MIT License',   # Again, pick a license
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',],)




"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from enum import IntFlag

class NodeClassMask(IntFlag):

    # <summary>
    # No classes are selected.
    # </summary>
    Unspecified = 0,

    #/ <summary>
    #/ The node is an object.
    #/ </summary>
    Object = 1,

    #/ <summary>
    #/ The node is a variable.
    #/ </summary>
    Variable = 2,

    #/ <summary>
    #/ The node is a method.
    #/ </summary>
    Method = 4,

    #/ <summary>
    #/ The node is an object type.
    #/ </summary>
    ObjectType = 8,

    #/ <summary>
    #/ The node is an variable type.
    #/ </summary>
    VariableType = 16,

    #/ <summary>
    #/ The node is a reference type.
    #/ </summary>
    ReferenceType = 32,

    #/ <summary>
    #/ The node is a data type.
    #/ </summary>
    DataType = 64,

    #/ <summary>
    #/ The node is a view.
    #/ </summary>
    View = 128,



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

import datetime
import re
from datetime import timedelta

from .aggregatedetails import AggregateType, AggregateDetails
from .itempathjson import ItemPathJSON
from .nodeattributes import NodeAttributes


class DatasetPen(object):
    """description of class"""
    def __init__(self, dataset, title, dataSourceName = None, penConfig = None):
        import clr
        self._dataset = dataset
        self._aggregateType = None
        self._dataSource = None
        if penConfig is None:
            self.setTitle(title)
            self._isRaw = True
            if dataSourceName is not None:
                self.setDataSource(dataSourceName)
        else:
            self._title = title
            self._penConfig = penConfig
            clr.AddReference('dll_references/EpmData')
            from Elipse.Epm.Common import RawByPeriod, TrendMode, DatasetField
            if not isinstance(penConfig, DatasetField):
                raise Exception("Invalid penConfig")
            nsInfo = penConfig.Identity.NamespaceIndex
            nodeId = penConfig.Identity.Numeric32Identifier
            identity = 'ns=' + str(nsInfo) + ';i=' + str(nodeId)
            itemPath = ItemPathJSON('OPCUA.NodeId', None, identity)
            result = dataset._connection._read([itemPath], [NodeAttributes.BrowseName.value] * len([itemPath])).items()
            dataSourceName = result[0][0].value.value[2:]
            self.setDataSource(dataSourceName)
            if isinstance(penConfig.Mode, RawByPeriod):
                self._isRaw = True
                self._aggregateType = None
            else:
                self._isRaw = False
                if isinstance(penConfig.Mode, TrendMode):
                    print(penConfig.Mode.SampleInterval.TotalMilliseconds)
                    self._getAggregateType(1, penConfig.Mode.SampleInterval.TotalMilliseconds)
                else:
                    self._getAggregateType(penConfig.Mode.CalculateType, penConfig.Mode.SampleInterval.TotalMilliseconds)


    @property
    def title(self):
        return self._title

    @property
    def dataSource(self):
        return self._dataSource

    @property
    def isRaw(self):
      return self._isRaw

    @property
    def aggregateType(self):
      return self._aggregateType

    def setTitle(self, title):
        if type(title) is str:
            if not title[0].isalpha():
                raise Exception("First character of the string argument must be a letter")
            pattern = r'^[A-Za-z0-9_!@#%&*.:-]+$'
            if re.search(pattern, title):
                self._title = title
            else:
                raise Exception("Invalid character on string argument")
        else:
            raise Exception("Argument must be a string")

    def setAggregationType(self, aggregateType, processInterval):
        if not isinstance(aggregateType, AggregateType):
            raise Exception("aggregateType argument must be of type AggregateType")
        if not isinstance(processInterval, datetime.timedelta):
            raise Exception("processInterval argument must be of type timedelta")
        self._aggregateType = AggregateDetails(processInterval, aggregateType)
        self._isRaw = False

    def setRawType(self):
        self._isRaw = True

    def setDataSource(self, name):
        self._dataSource = self._dataset._connection.getDataObjects(name)[name]

    def _getAggregateType(self, aggregateId, processInterval):
        import clr
        clr.AddReference('dll_references/EpmData')
        from Elipse.Epm.Common import AggregateType as at
        if aggregateId == at.Trend:
            aggregateType = AggregateType.Trend
        elif aggregateId == at.Interpolative:
            aggregateType = AggregateType.Interpolative
        elif aggregateId == at.Average:
            aggregateType = AggregateType.Average
        elif aggregateId == at.Total:
            aggregateType = AggregateType.Total
        elif aggregateId == at.Count:
            aggregateType = AggregateType.Count
        elif aggregateId == at.Minimum:
            aggregateType = AggregateType.Minimum
        elif aggregateId == at.Maximum:
            aggregateType = AggregateType.Maximum
        elif aggregateId == at.MinimumActualTime:
            aggregateType = AggregateType.MinimumActualTime
        elif aggregateId == at.MaximumActualTime:
            aggregateType = AggregateType.MaximumActualTime
        elif aggregateId == at.Range:
            aggregateType = AggregateType.Range
        elif aggregateId == at.Delta:
            aggregateType = AggregateType.Delta
        elif aggregateId == at.TimeAverage:
            aggregateType = AggregateType.TimeAverage
        elif aggregateId == at.AnnotationCount:
            aggregateType = AggregateType.AnnotationCount
        elif aggregateId == at.DurationInStateZero:
            aggregateType = AggregateType.DurationInStateZero
        elif aggregateId == at.DurationInStateNonZero:
            aggregateType = AggregateType.DurationInStateNonZero
        elif aggregateId == at.NumberOfTransitions:
            aggregateType = AggregateType.NumberOfTransitions
        elif aggregateId == at.Start:
            aggregateType = AggregateType.Start
        elif aggregateId == at.End:
            aggregateType = AggregateType.End
        elif aggregateId == at.DurationGood:
            aggregateType = AggregateType.DurationGood
        elif aggregateId == at.DurationBad:
            aggregateType = AggregateType.DurationBad
        elif aggregateId == at.PercentGood:
            aggregateType = AggregateType.PercentGood
        elif aggregateId == at.PercentBad:
            aggregateType = AggregateType.PercentBad
        elif aggregateId == at.WorstQuality:
            aggregateType = AggregateType.WorstQuality
        elif aggregateId == at.PercentInStateZero:
            aggregateType = AggregateType.PercentInStateZero
        elif aggregateId == at.PercentInStateNonZero:
            aggregateType = AggregateType.PercentInStateNonZero
        elif aggregateId == at.TimeAverage2:
            aggregateType = AggregateType.TimeAverage2
        elif aggregateId == at.Total2:
            aggregateType = AggregateType.Total2
        elif aggregateId == at.Minimum2:
            aggregateType = AggregateType.Minimum2
        elif aggregateId == at.Maximum2:
            aggregateType = AggregateType.Maximum2
        elif aggregateId == at.MinimumActualTime2:
            aggregateType = AggregateType.MinimumActualTime2
        elif aggregateId == at.MaximumActualTime2:
            aggregateType = AggregateType.MaximumActualTime2
        elif aggregateId == at.Range2:
            aggregateType = AggregateType.Range2
        elif aggregateId == at.StartBounds:
            aggregateType = AggregateType.StartBounds
        elif aggregateId == at.EndBounds:
            aggregateType = AggregateType.EndBounds
        elif aggregateId == at.DeltaBounds:
            aggregateType = AggregateType.DeltaBounds
        elif aggregateId == at.WorstQuality2:
            aggregateType = AggregateType.WorstQuality2
        elif aggregateId == at.StandardDeviationPopulation:
            aggregateType = AggregateType.StandardDeviationPopulation
        elif aggregateId == at.VariancePopulation:
            aggregateType = AggregateType.VariancePopulation
        elif aggregateId == at.StandardDeviationSample:
            aggregateType = AggregateType.StandardDeviationSample
        elif aggregateId == at.VarianceSample:
            aggregateType = AggregateType.VarianceSample
        else:
            raise Exception("Aggregation type error")

        processInterval = timedelta(milliseconds=processInterval)

        self._aggregateType = AggregateDetails(processInterval, aggregateType)

    def _getAggregateId(self):
        import clr
        clr.AddReference('dll_references/EpmData')
        from Elipse.Epm.Common import AggregateType as at
        aggregateType = self.aggregateType.type
        if aggregateType == AggregateType.Trend.name:
            aggregateId = at.Trend
        elif aggregateType == AggregateType.Interpolative.name:
            aggregateId = at.Interpolative
        elif aggregateType == AggregateType.Average.name:
            aggregateId = at.Average
        elif aggregateType == AggregateType.Total.name:
            aggregateId = at.Total
        elif aggregateType == AggregateType.Count.name:
            aggregateId = at.Count
        elif aggregateType == AggregateType.Minimum.name:
            aggregateId = at.Minimum
        elif aggregateType == AggregateType.Maximum.name:
            aggregateId = at.Maximum
        elif aggregateType == AggregateType.MinimumActualTime.name:
            aggregateId = at.MinimumActualTime
        elif aggregateType == AggregateType.MaximumActualTime.name:
            aggregateId = at.MaximumActualTime
        elif aggregateType == AggregateType.Range.name:
            aggregateId = at.Range
        elif aggregateType == AggregateType.Delta.name:
            aggregateId = at.Delta
        elif aggregateType == AggregateType.TimeAverage.name:
            aggregateId = at.TimeAverage
        elif aggregateType == AggregateType.AnnotationCount.name:
            aggregateId = at.AnnotationCount
        elif aggregateType == AggregateType.DurationInStateZero.name:
            aggregateId = at.DurationInStateZero
        elif aggregateType == AggregateType.DurationInStateNonZero.name:
            aggregateId = at.DurationInStateNonZero
        elif aggregateType == AggregateType.NumberOfTransitions.name:
            aggregateId = at.NumberOfTransitions
        elif aggregateType == AggregateType.Start.name:
            aggregateId = at.Start
        elif aggregateType == AggregateType.End.name:
            aggregateId = at.End
        elif aggregateType == AggregateType.DurationGood.name:
            aggregateId = at.DurationGood
        elif aggregateType == AggregateType.DurationBad.name:
            aggregateId = at.DurationBad
        elif aggregateType == AggregateType.PercentGood.name:
            aggregateId = at.PercentGood
        elif aggregateType == AggregateType.PercentBad.name:
            aggregateId = at.PercentBad
        elif aggregateType == AggregateType.WorstQuality.name:
            aggregateId = at.WorstQuality
        elif aggregateType == AggregateType.PercentInStateZero.name:
            aggregateId = at.PercentInStateZero
        elif aggregateType == AggregateType.PercentInStateNonZero.name:
            aggregateId = at.PercentInStateNonZero
        elif aggregateType == AggregateType.TimeAverage2.name:
            aggregateId = at.TimeAverage2
        elif aggregateType == AggregateType.Total2.name:
            aggregateId = at.Total2
        elif aggregateType == AggregateType.Minimum2.name:
            aggregateId = at.Minimum2
        elif aggregateType == AggregateType.Maximum2.name:
            aggregateId = at.Maximum2
        elif aggregateType == AggregateType.MinimumActualTime2.name:
            aggregateId = at.MinimumActualTime2
        elif aggregateType == AggregateType.MaximumActualTime2.name:
            aggregateId = at.MaximumActualTime2
        elif aggregateType == AggregateType.Range2.name:
            aggregateId = at.Range2
        elif aggregateType == AggregateType.StartBounds.name:
            aggregateId = at.StartBounds
        elif aggregateType == AggregateType.EndBounds.name:
            aggregateId = at.EndBounds
        elif aggregateType == AggregateType.DeltaBounds.name:
            aggregateId = at.DeltaBounds
        elif aggregateType == AggregateType.WorstQuality2.name:
            aggregateId = at.WorstQuality2
        elif aggregateType == AggregateType.StandardDeviationPopulation.name:
            aggregateId = at.StandardDeviationPopulation
        elif aggregateType == AggregateType.VariancePopulation.name:
            aggregateId = at.VariancePopulation
        elif aggregateType == AggregateType.StandardDeviationSample.name:
            aggregateId = at.StandardDeviationSample
        elif aggregateType == AggregateType.VarianceSample.name:
            aggregateId = at.VarianceSample
        else:
            raise Exception("Aggregation type error")

        return aggregateId



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""
import base64
import json
import os


class CustomTypeDefinition:
    """description of class"""

    def __init__(self, epmConnection, name, icon=None, aliasProperties=None, simpleProperties=None,
                 objectProperties=None, placeHolderTypes=None):
        self._epmConnection = epmConnection
        self._name = name

        if icon:
            self._icon = icon
        else:
            self._icon = ""

        if aliasProperties:
            self._aliasProperties = aliasProperties
        else:
            self._aliasProperties = []

        if simpleProperties:
            self._simpleProperties = simpleProperties
        else:
            self._simpleProperties = []

        if objectProperties:
            self._objectProperties = objectProperties
        else:
            self._objectProperties = []

        if placeHolderTypes:
            self._placeHolderTypes = placeHolderTypes
        else:
            self._placeHolderTypes = []

    def __str__(self):
        stringDefinition = "Custom Type Definition"
        stringDefinition += "\nname: " + self._name

        if self._icon == "" or self._icon == self.defaultIcon:
            icon = "Default"
        else:
            icon = self._icon

        stringDefinition += "\nicon: " + icon

        if self._aliasProperties:
            for alias in self._aliasProperties:
                stringDefinition += "\naliasProperties[" + str(self._aliasProperties.index(alias)) + "]: " + alias

        if self._simpleProperties:
            for simple in self._simpleProperties:
                if simple.initialValue:
                    initialValue = str(simple.initialValue)
                else:
                    initialValue = "None"
                stringDefinition += "\nsimpleProperties[" + str(self._simpleProperties.index(simple)) + "]: name = " + \
                                    simple.name + ", initialValue = " + initialValue

        if self._objectProperties:
            for obj in self._objectProperties:
                stringDefinition += "\nobjectProperties[" + str(self._objectProperties.index(obj)) + "]: name = " + \
                                    obj.name + ", customType = " + obj.customType

        if self._placeHolderTypes:
            for placeHolder in self._placeHolderTypes:
                stringDefinition += "\nplaceHolderTypes[" + str(self._placeHolderTypes.index(placeHolder)) + "]: " + placeHolder

        return stringDefinition

    @property
    def name(self):
        return self._name

    @property
    def icon(self):
        return self._icon

    @property
    def aliasProperties(self):
        return self._aliasProperties

    @property
    def simpleProperties(self):
        return self._simpleProperties

    @property
    def objectProperties(self):
        return self._objectProperties

    @property
    def placeHolderTypes(self):
        return self._placeHolderTypes

    @property
    def defaultIcon(self):
        return "iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7" \
               "DAcdvqGQAAAJ7SURBVDhPnVNLa1NREJ6btDdptfZhWkiELlx05cIui+5ERNwrFOxPcOUPEESK4EIQd7pxI7qptFgFH7gSt4LQFpWKAW" \
               "uSm7R53vO6c/zmJqRod174mMO5M9/MfHMmKJVK9O83f38zy6q7EAS06Nmf8rH66C19+rW6kgxchl9KsPTk/Xwxn7usmM82bbLYMO6MT" \
               "pJjbB0JvDXE2tbIuJccx+uu0X5Te3yzkxKcf7WVu1qaKp8Is7N1nVBkHNUAsfuAs5YY1jtLHjbnEeWsirXZULuV5eDcxtbMheJEvZQf" \
               "pcmRLMXMVFEgQOaadmS8p8lMQF5raseaap2YWhpkSULx1s9isLT2JXt6Oh8Xx8LRQjhCc7kROg4inTBFsaFvjRZtR01q9TQyOyJGCSB" \
               "lY1iXo/FUg4VHbytT+XBuFsGFMEulXEjGGHq+U0b/KN9YImT0DgAxoUq2ttF8PXkyI0LAOYpQ3m5P0feuoR84GzizNuQVAMu4YyVQ/X" \
               "ttq2rnOqUEEKfGiaMYve+h7K9tjYSYGERj9M4gQckpREgvwlpTkdg+gXORN1IifgANBFgZ4SAAzv3AwXlwXx0SYNZRnxVlixNIerAk/" \
               "YtwjlObAj4oHwT2kAAvbbfviB+w8ng60rcEpOKJBdIqkF18rTtswdUPHnA3/iAVSCUyrja0kGBKlQfSFiWJnOVxmc9DgsrDGz3zO7rC" \
               "ne47GZW00YUOMnOPkWEfMHqQYYQY6w6wnLSSdYn9a5kKK6vjYXH2RTAxfnEaWat70aGASm8j8A5aeHqweW+4VEe2cebarbHcfHFtNJO" \
               "5pJod4p7axgLdxlN+tr9+9+g2ejzL//+I/gCSwRoBoUN0GgAAAABJRU5ErkJggg=="

    # Public Methods

    def updateIcon(self, pathName):
        with open(pathName, "rb") as imageFile:
            icon = base64.b64encode(imageFile.read()).decode("utf-8")
            if self._icon == icon:
                return
            self._icon = icon

    def removeIcon(self):
        icon = ""
        if self._icon == icon:
            return
        self._icon = icon

    def addAliasProperty(self, aliasName):
        if len(self._aliasProperties) == 0:
            self._aliasProperties.append(aliasName)
        else:
            names = []
            for name in self._aliasProperties:
                names.append(name)

            if aliasName not in names:
                self._aliasProperties.append(aliasName)
            else:
                raise Exception("Name already exists")

    def removeAliasProperty(self, aliasName):
        names = []
        for name in self._aliasProperties:
            names.append(name)

        if aliasName in names:
            self._aliasProperties.remove(aliasName)
        else:
            raise Exception("Alias Property not found")

    def addSimpleProperty(self, simpleName, initialValue = None):
        if len(self._simpleProperties) == 0:
            newSimpleProperty = SimpleProperty(simpleName, initialValue)
        else:
            names = []
            for item in self._simpleProperties:
                names.append(item.name)
            if simpleName not in names:
                newSimpleProperty = SimpleProperty(simpleName, initialValue)
            else:
                raise Exception("Name already exists")

        self._simpleProperties.append(newSimpleProperty)

    def removeSimpleProperty(self, simpleName):
        names = []
        for item in self._simpleProperties:
            names.append(item.name)
        if simpleName in names:
            self._simpleProperties.pop(names.index(simpleName))
        else:
            raise Exception("Simple Property not found")

    def addObjectProperty(self, objectName, customType):
        if customType not in self._epmConnection.getAllCustomTypes():
            raise Exception("Object Property type does not exist")

        if len(self._objectProperties) == 0:
            newObjectProperty = ObjectProperty(objectName, customType)
        else:
            names = []
            for item in self._objectProperties:
                names.append(item.name)
            if objectName not in names:
                newObjectProperty = ObjectProperty(objectName, customType)
            else:
                raise Exception("Name already exists")

        self._objectProperties.append(newObjectProperty)

    def removeObjectProperty(self, objectName):
        names = []
        for item in self._objectProperties:
            names.append(item.name)
        if objectName in names:
            self._objectProperties.pop(names.index(objectName))
        else:
            raise Exception("Object Property not found")

    def addPlaceHolderType(self, customTypeName):
        if customTypeName in self._placeHolderTypes:
            raise Exception("Type already in a PlaceHolder")

        if customTypeName in self._epmConnection.getAllCustomTypes():
            self._placeHolderTypes.append(customTypeName)
        else:
            raise Exception("PlaceHolder type does not exist")

    def removePlaceHolderType(self, customTypeName):
        if customTypeName in self._placeHolderTypes:
            self._placeHolderTypes.remove(customTypeName)
        else:
            raise Exception("PlaceHolder not found")

    def save(self):
        self._epmConnection.updateCustomType(self.exportJSON())

    def delete(self):
        self._epmConnection.deleteCustomType(self._name)

    def exportJSON(self, fileName = None, pathName = None):
        aliasPropertiesList = []
        for aliasName in self._aliasProperties:
            aliasPropertiesList.append({'name': aliasName})

        simplePropertiesList = []
        for simpleProperty in self._simpleProperties:
            simplePropertiesList.append({'name': simpleProperty.name, 'initialValue': simpleProperty.initialValue})

        objectPropertiesList = []
        for objectProperty in self._objectProperties:
            objectPropertiesList.append({'name': objectProperty.name, 'type': objectProperty.customType})

        placeHolderPropertiesList = []
        for placeHolderType in self._placeHolderTypes:
            placeHolderPropertiesList.append({'type': placeHolderType})

        propertiesJSON = {'name': self.name, 'icon': self.icon, 'aliasProperties': aliasPropertiesList,
                          'simpleProperties': simplePropertiesList, 'objectProperties': objectPropertiesList,
                          'placeHolderProperties': placeHolderPropertiesList}

        if fileName is not None:
            if pathName is None:
                pathName = os.getcwd()
            fileName = fileName + '.json'
            with open(os.path.join(pathName, fileName), 'w') as outfile:
                json.dump(propertiesJSON, outfile)

        return propertiesJSON


class SimpleProperty:

    def __init__(self, name, initialValue = None):
        self._name = name
        self._initialValue = initialValue

    @property
    def name(self):
        return self._name

    @property
    def initialValue(self):
        return self._initialValue


class ObjectProperty:

    def __init__(self, name, customType):
        self._name = name
        self._customType = customType

    @property
    def name(self):
        return self._name

    @property
    def customType(self):
        return self._customType


class CustomTypeAlreadyExistsException(Exception):
    pass


class InvalidCustomTypeNameException(Exception):
    pass


class CustomTypeDependenciesException(Exception):
    pass


class InvalidIconException(Exception):
    pass


class DuplicatedPropertiesNamesException(Exception):
    pass


class DuplicatedPropertiesTypeException(Exception):
    pass


class MissingPropertyNameException(Exception):
    pass


class InvalidPropertyTypeException(Exception):
    pass


class InvalidPropertyNameException(Exception):
    pass



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class QueryPeriod(object):
    """description of class"""

    def __init__(self, start, end):
        self._start = start
        self._end = end

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        self._end = value



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class HistoryResultItemModelJSON(object):
    """description of class"""
    def __init__(self, continuationPoint, dataValues):
        self._continuationPoint = continuationPoint
        self._dataValues = dataValues





"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

from .epmobject import EpmObject

class EpmModelObject(EpmObject):
    """description of class"""

    def __init__(self, epmConnection, identity, path, name, type = ''):
        super().__init__(epmConnection, identity, path, name, type)

    # Public Methods

    def addInstances(self, names, objTypes):
        splitPath = self._path.split("/")

        if splitPath[0] == "":
            addPath = "/".join(splitPath[3:])
        else:
            addPath = "/".join(splitPath[2:])

        return self._epmConnection.addInstancesEpmModel(addPath, names, objTypes)

    def removeInstance(self, name):
        splitPath = self._path.split("/")

        if splitPath[0] == "":
            removePath = "/".join(splitPath[3:]) + "/" + name
        else:
            removePath = "/".join(splitPath[2:]) + "/" + name

        if removePath[0] == "/":
            removePath = removePath[1:]

        self._epmConnection.removeInstanceEpmModel(removePath)


    def setBindedVariables(self, aliasProperties, variablesNames):
        splitPath = self._path.split("/")

        if splitPath[0] == "":
            objectPath = "/".join(splitPath[3:])
        else:
            objectPath = "/".join(splitPath[2:])

        self._epmConnection.setBindedVariables(objectPath, aliasProperties, variablesNames, True)


class ObjectDependenciesException(Exception):
    pass


class InstanceNameDuplicatedException(Exception):
    pass


class InvalidObjectNameException(Exception):
    pass


class InvalidSourceVariableException(Exception):
    pass


class InvalidObjectTypeException(Exception):
    pass


class InvalidObjectPropertyException(Exception):
    pass



"""
Elipse Plant Manager - EPM Web API
Copyright (C) 2018 Elipse Software.
Distributed under the MIT License.
(See accompanying file LICENSE.txt or copy at http://opensource.org/licenses/MIT)
"""

class ReadItemModelJSON(object):
  
  def __init__(self, path, attributeId):
    self._path = path
    self._attributeId = attributeId

  def toDict(self):

    return { 'path': self._path.toDict(), 
             'attributeId': self._attributeId }

